#!/bin/env python
# Copyright (c) 2013, 2018 National Technology and Engineering Solutions of Sandia, LLC . Under the terms of Contract
# DE-NA0003525 with National Technology and Engineering Solutions of Sandia, LLC, the U.S. Government
# retains certain rights in this software.

"""Demonstrates uploading data to Slycat Web Server to compute a CCA model.

This script computes a Slycat cca model using a set of semi-random observations.  Use
this script as a starting-point for uploading your own data to a CCA model.

A Slycat CCA model requires the following data, which you will have to provide
in your own scripts:

    data-table        A single 1D array containing M input observations with N features (array attributes).
"""

import numpy
import slycat.web.client

# parse arguments
def parse_arguments(list_input=None):

  # get and parse arguments for creating CCA model
  parser = slycat.web.client.ArgumentParser(description="Create CCA model from randomly generated data.")
  parser.add_argument("--column-prefix", default="a", help="Column prefix.  Default: %(default)s")
  parser.add_argument("--constant-input-count", type=int, default=0, help="Number of input columns to make constant.  Default: %(default)s")
  parser.add_argument("--constant-output-count", type=int, default=0, help="Number of output columns to make constant.  Default: %(default)s")
  parser.add_argument("--duplicate-input-count", type=int, default=0, help="Number of input columns to duplicate.  Default: %(default)s")
  parser.add_argument("--duplicate-output-count", type=int, default=0, help="Number of output columns to duplicate.  Default: %(default)s")
  parser.add_argument("--input-count", type=int, default=3, help="Input column count.  Default: %(default)s")
  parser.add_argument("--marking", default="", help="Marking type.  Default: %(default)s")
  parser.add_argument("--model-name", default="Sample CCA Model", help="New model name.  Default: %(default)s")
  parser.add_argument("--output-count", type=int, default=3, help="Output column count.  Default: %(default)s")
  parser.add_argument("--project-name", default="Sample CCA Project", help="New project name.  Default: %(default)s")
  parser.add_argument("--row-count", type=int, default=100, help="Row count.  Default: %(default)s")
  parser.add_argument("--seed", type=int, default=12345, help="Random seed.  Default: %(default)s")
  parser.add_argument("--unused-count", type=int, default=3, help="Unused column count.  Default: %(default)s")
  arguments = parser.parse_args(list_input)

  if arguments.input_count < 1:
    raise Exception("Input count must be greater-than zero.")
  if arguments.output_count < 1:
    raise Exception("Output count must be greater-than zero.")
  if arguments.constant_input_count > arguments.input_count:
    raise Exception("Constant input count must be less than input count.")
  if arguments.constant_output_count > arguments.output_count:
    raise Exception("Constant output count must be less than output count.")
  if arguments.duplicate_input_count >= arguments.input_count:
    raise Exception("Duplicate input count must be less than input count.")
  if arguments.duplicate_output_count >= arguments.output_count:
    raise Exception("Duplicate output count must be less than output count.")

  return arguments

# create CCA model
def main(arguments, connection):

  total_columns = arguments.input_count + arguments.output_count + arguments.unused_count

  # Create some random data using a gaussian distribution.
  numpy.random.seed(arguments.seed)
  data = numpy.random.normal(size=(arguments.row_count, total_columns))

  # Force a somewhat-linear relationship between the inputs and outputs.
  for i in range(arguments.input_count, arguments.input_count + min(arguments.input_count, arguments.output_count)):
    data[:, i] = data[:, 0] ** i

  # Optionally make some columns constant.
  for i in range(arguments.constant_input_count):
    data[:,i] = data[0,i]
  for i in range(arguments.input_count, arguments.input_count + arguments.constant_output_count):
    data[:,i] = data[0,i]

  # Optionally duplicate some columns to create rank-deficient data.
  for i in range(1, 1 + arguments.duplicate_input_count):
    data[:,i] = data[:,0]
  for i in range(1 + arguments.input_count, 1 + arguments.input_count + arguments.duplicate_output_count):
    data[:,i] = data[:, arguments.input_count]

  # Create a new project to contain our model.
  pid = connection.find_or_create_project(arguments.project_name)

  # Create the new, empty model.
  mid = connection.post_project_models(pid, "cca", arguments.model_name, arguments.marking)

  # Upload our observations as "data-table".
  connection.put_model_arrayset(mid, "data-table")

  # Start our single "data-table" array.
  dimensions = [dict(name="row", end=arguments.row_count)]
  attributes = [dict(name="%s%s" % (arguments.column_prefix, column), type="float64") for column in range(total_columns)]
  connection.put_model_arrayset_array(mid, "data-table", 0, dimensions, attributes)

  # Upload data into the array.
  for i in range(total_columns):
    connection.put_model_arrayset_data(mid, "data-table", "0/%s/..." % i, [data.T[i]])

  # Store the remaining parameters.
  connection.put_model_parameter(mid, "input-columns", list(range(0, arguments.input_count)))
  connection.put_model_parameter(mid, "output-columns", 
    list(range(arguments.input_count, arguments.input_count + arguments.output_count)))
  connection.put_model_parameter(mid, "scale-inputs", False)

  # Signal that we're done uploading data to the model.  This lets Slycat Web
  # Server know that it can start computation.
  connection.post_model_finish(mid)
  # Wait until the model is ready.
  connection.join_model(mid)

  # Supply the user with a direct link to the new model.
  host = arguments.host
  if arguments.port:
      host = host + ":" + arguments.port
  slycat.web.client.log.info("Your new model is located at %s/models/%s" % (host, mid))

  return mid

# command line entry point
if __name__ == "__main__":
  
  # set up parser and get arguments
  arguments = parse_arguments()

  # connect and get projects
  connection = slycat.web.client.connect(arguments)

  # list projects
  main(arguments, connection)



# Copyright (c) 2013, 2018 National Technology and Engineering Solutions of Sandia, LLC. 
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering 
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this software.

# Creates dial-a-cluster run chart models by processing TDMS formatted 
# files and creating a dial-a-cluster model.  The model can be saved
# in the dac genric format and/or pushed directly to a Slycat server.

# S. Martin
# 8/30/2022

# connection to Slycat server
import slycat.web.client

# DAC upload
import slycat.web.client.dac_gen as dac_gen

# file name manipulation
import os
import fnmatch

# manipulating command line arguments
from argparse import Namespace

# processing tables
import numpy as np

# compute PCA with sklearn
from sklearn.decomposition import PCA

# reading tdms files
import nptdms
import pandas as pd

# output zip file
from zipfile import ZipFile

# the tdms file types allowed
TDMS_MATCHES = ['Factory_Trigger', 'Acceptance_Trigger', 'Pulse_Life', 
                'Extended_Pulse_Life', 'Probe_Age']

# the group run chart variables allowed
RUN_CHART_MATCHES = ['TAD', 'Rp', 'Ip', 'DBV']
RUN_CHART_UNITS = ['nsec', 'kOhms', 'Amps', 'V']

# TDMS error handling
class TDMSUploadError (Exception):

    # exception for TDMS upload problems
    def __init__(self, message):
        self.message = message

# parse test data subdirectory
def parse_run_chart (run_chart):

    # default return is None
    run_chart_id = None

    # split run_chart string on "_" to get identifiers
    run_chart_id = run_chart.split("_")

    # there should be four identifiers
    if len(run_chart_id) != 4:
        run_chart_id = None
    
    return run_chart_id

# organize tdms files based on command line inputs
def catalog_tdms_files (arguments, log):

    # gather data directories for given part number
    root_dir = arguments.input_data_dir
    part_num_match = arguments.part_num_match

    # check that root dir exists
    if not os.path.isdir(root_dir):
        raise TDMSUploadError("Input data directory does not exist.  Please provide " +
              "a different directory and try again.")
    
    # look for directories containing data for part number provided
    root_subdirs = os.listdir(root_dir)
    part_subdirs = fnmatch.filter(root_subdirs, part_num_match)
    
    # are there any subdirectories?
    if len(part_subdirs) == 0:
        raise TDMSUploadError('Could not find any subdirectories for part number matching "' +
              part_num_match + '".')

    # put subdirectories in order
    part_subdirs.sort()

    # get list of .tdms file matches
    tdms_matches = TDMS_MATCHES
    if arguments.tdms_file_matches is not None:
        tdms_matches = arguments.tdms_file_matches
    tdms_matches = [file_match.replace('_', ' ') for file_match in tdms_matches]

    # report on tdms file being used
    log("Using .tdms file matches in: " + str(tdms_matches))

    # look through each subdirectory for run chart data
    metadata = []
    for subdir in part_subdirs:

        # look for test data directory
        test_data_dir = os.path.join(root_dir, subdir, 'Test Data')

        # skip if run_chart_dir is not an actual directory
        if not os.path.isdir(test_data_dir):
            log('Skipping "' + test_data_dir + '" because it''s not a directory.')
            continue

        # each of the subdirectories of run_chart_dir will be a row in the 
        # metadata table for the run chart model
        possible_run_chart_dirs = os.listdir(test_data_dir)

        # sort according to data ids
        possible_run_chart_dirs.sort()

        # check that we found data directories
        if len(possible_run_chart_dirs) == 0:
            raise TDMSUploadError('No data subdirectories found for part number match "' + 
                  part_num_match + '".')

        # check that data directories conform to expected format
        for run_chart_dir in possible_run_chart_dirs:

            # check that directory is expected format for a run chart
            run_chart_ids = parse_run_chart(run_chart_dir)

            # if it's not a run chart then skip it
            if run_chart_ids is None:
                log('Skipping run chart subdirectory "' + run_chart_dir + 
                    '" because it does not conform to "part_lot_batch_serial" string format.')
                continue
        
            # find .tdms files in run chart directory
            run_chart_files = os.listdir(os.path.join(test_data_dir, run_chart_dir))
            run_chart_tdms_files = [run_chart_file for run_chart_file in run_chart_files 
                                    if run_chart_file.endswith('.tdms')]
        
            # check that we have .tdms files
            if len(run_chart_tdms_files) == 0:
                log ('Skipping subdirectory "' + run_chart_ids[0] + '" -- does not ' +
                      'contain any TDMS files.')
                continue

            # screen for .tdms files according to argument parameters
            run_chart_tdms_matches = []
            run_chart_tdms_types = []
            for run_chart_file in run_chart_tdms_files:
                for match in tdms_matches:
                    if match in run_chart_file:

                        # check that .tdms file has matching part_lot_batch_sn
                        if not run_chart_file.startswith(
                            run_chart_ids[0] + "_" + 
                            run_chart_ids[1] + "_" + 
                            run_chart_ids[2] + "_" + 
                            run_chart_ids[3]):
                            continue
                        
                        # keep track of files and types
                        run_chart_tdms_types.append(match)
                        run_chart_tdms_matches.append(run_chart_file)

            # check that we have matching .tdms files
            if len(run_chart_tdms_matches) == 0:
                log ('Skipping subdirectory "' + run_chart_ids[0] + '" -- does not ' +
                      'contain any TDMS files with file matches.')
                continue

            # store in metadata structure
            metadata.append({"part": run_chart_ids[0],
                             "lot": run_chart_ids[1],
                             "batch": run_chart_ids[2],
                             "sn": run_chart_ids[3],
                             "source": os.path.join(os.path.abspath(test_data_dir), 
                                                    run_chart_dir),
                             "tdms_files": run_chart_tdms_matches,
                             "tdms_types": run_chart_tdms_types})

    # check that files were found
    if metadata == []:
        raise TDMSUploadError("No TDMS files matching selection criterion were found.")

    # get common set of tdms types
    common_tdms_types = set(metadata[0]["tdms_types"])
    intersection_flag = False
    for row in range(len(metadata)):

        # set intersection
        common_tdms_types = common_tdms_types.intersection(set(metadata[row]["tdms_types"]))

        # set flag if intersection reduced number of files
        if set(metadata[row]["tdms_types"]) != common_tdms_types:
            intersection_flag = True

    # remove files with uncommon types
    if intersection_flag:
        for row in range(len(metadata)):
            for tdms_ind in reversed(range(len(metadata[row]["tdms_types"]))):
                if metadata[row]["tdms_types"][tdms_ind] not in common_tdms_types:

                    # warn user we are ignoring this file
                    log('Ignoring file "' + metadata[row]["tdms_files"][tdms_ind] +
                        '" because type "' + metadata[row]["tdms_types"][tdms_ind] +
                        '" is not present in all runs.')

                    # remove files from catalog
                    metadata[row]["tdms_types"].pop(tdms_ind)
                    metadata[row]["tdms_files"].pop(tdms_ind)

    # sort catalog by tdms type
    for row in range(len(metadata)):

        # get tmds types and files
        row_tdms_types = metadata[row]["tdms_types"]
        row_tdms_files = metadata[row]["tdms_files"]

        # get indices of sorted tdms types
        tdms_inds = np.argsort(row_tdms_types)

        # sort and re-store types and files
        metadata[row]["tdms_types"] = list(np.asarray(row_tdms_types)[tdms_inds])
        metadata[row]["tdms_files"] = list(np.asarray(row_tdms_files)[tdms_inds])

    return metadata 

# read all tdms run chart data
def read_tdms_files(metadata, run_chart_matches, log):

    # go through each row in the table
    skipped_rows = np.zeros(len(metadata))
    for row in range(len(metadata)):

        root_properties = []
        run_charts = []
        run_chart_headers = []
        tdms_types = []
        for tdms_file_ind in range(len(metadata[row]["tdms_files"])):

            # report progress
            log('Reading .tdms file: "' + metadata[row]["tdms_files"][tdms_file_ind] + '".')

            # save run chart type
            tdms_types.append(metadata[row]["tdms_types"][tdms_file_ind])

            # open tdms file
            tdms_file_name = metadata[row]["tdms_files"][tdms_file_ind]
            tdms_file_path = os.path.join(metadata[row]["source"], tdms_file_name)
            with nptdms.TdmsFile.open(tdms_file_path) as tdms_file:

                # save the root properties for the .tdms file
                root_properties.append(pd.DataFrame(tdms_file.properties, index=[0]))

                # get group properties
                group_header = None
                group_rows = []
                for group in tdms_file.groups():

                    # check that header is the same
                    if group_header is None:
                        group_header = list(group.properties.keys())

                    # this should never happen
                    elif group_header != list(group.properties.keys()):
                        TDMSUploadError('Group headers in file "' + tdms_file_name + 
                                        '" do not match.')

                    # keep track of table values
                    group_rows.append(list(group.properties.values()))

                # create pandas dataframe for group properies
                group_df = pd.DataFrame(data=group_rows, columns=group_header)

                # find numeric columns
                numeric_group_df = group_df.select_dtypes('float64')

                # select headers that with matching run charts
                header_matches = []
                rc_matches = []
                for header in numeric_group_df.head():
                    for run_chart in run_chart_matches:
                        if run_chart in header:
                            header_matches.append(header)
                            rc_matches.append(run_chart)

                # check that run charts have enough data
                if any(pd.notna(numeric_group_df[header_matches]).sum()==0):
                    log('Incomplete run chart data found for "' + tdms_file_path + '" -- ' +
                        'skipping directory "' + metadata[row]["source"] + '".')
                    skipped_rows[row] = 1

                # get indices of sorted tdms types
                rc_inds = np.argsort(rc_matches)

                # re-order run charts and headers alphabetically
                rc_matches = list(np.asarray(rc_matches)[rc_inds])
                header_matches = list(np.asarray(header_matches)[rc_inds])

                # check that run chart data is non-empty
                if rc_matches == []:
                    log('No run chart data found for "' + tdms_file_path + '" --' +
                        'skipping directory "' + metadata[row]["source"] +'".')
                    skipped_rows[row] = 1
                
                else:
                    # get sub dataframe with matching headers
                    run_charts.append(numeric_group_df[header_matches])

                    # also save run chart header matches
                    run_chart_headers.append(rc_matches)

        # add to metadata
        metadata[row]["root_properties"] = root_properties
        metadata[row]["run_charts"] = run_charts
        metadata[row]["run_chart_headers"] = run_chart_headers
    
    # remove skipped rows
    for row in reversed(range(len(metadata))):
        if skipped_rows[row]:
            metadata.pop(row)

    # check that data is still non-empty
    if metadata == []:
        raise TDMSUploadError("Too many run charts were missing.  " + 
            "No data remaining to be processed.")

    # go through metadata and check that run charts are the same, per tdms_type
    run_charts = metadata[0]["run_chart_headers"]
    for row in range(len(metadata)):
        for tdms_type in range(len(metadata[row]["tdms_types"])):
            if metadata[row]["run_chart_headers"][tdms_type] != run_charts[tdms_type]:
                raise TDMSUploadError('Found non-matching run chart data in directory"' +
                    metadata[row]["source"] + '", can not continue.')

    return metadata

# fill in metadata table
def read_metadata_table (metadata, arguments, log):

    # go through once to check that column names are the same
    root_head = list(metadata[0]["root_properties"][0].columns)
    for i in range(len(metadata)):
        root_props = metadata[i]["root_properties"]
        
        # check if root properties have same header
        for j in range(len(root_props)):
        
            root_props_j = list(root_props[j].columns)
            if root_props_j != root_head:
                
                # switch to longer root header
                if len(root_props_j) > len(root_head):
                    root_head = root_props_j
                    
    # remove shorter root headers?
    for i in reversed(range(len(metadata))):
        root_props = metadata[i]["root_properties"]

        # is it a short root header?
        delete_i = False
        for j in range(len(root_props)):
        
            root_props_j = list(root_props[j].columns)
            if root_props_j != root_head:
               
                # fill in missing columns with NaNs
                for k in range(len(root_head)):
                    if root_head[k] not in root_props_j:
                        root_props[j][root_head[k]] = np.nan

                log ('Found TDMS file with inconsistent root property header: "' +
                     metadata[i]["source"] + '" for "' + metadata[i]["tdms_types"][j] + 
                    '" -- using NaN values for missing columns.')
                delete_i = True
        
        # delete entry for any short headers
        # if delete_i:
        #     metadata.pop(i)

    # go through and get constant values
    constants = np.ones(len(root_head))
    for i in range(len(metadata)):
        root_props = metadata[i]["root_properties"]

        # check for constant values/nans
        for k in range(len(root_head)):
        
            const_val = root_props[0].iloc[0][root_head[k]]     
            for j in range(len(root_props)):
            
                # check for constant value
                curr_val = root_props[j].iloc[0][root_head[k]]
                if curr_val != const_val:
                    constants[k] = 0
                    
    
    # collect constant/variable/nan columns
    const_cols = list(np.asarray(root_head)[np.where(constants)[0]])
    var_cols = list(np.asarray(root_head)[np.where(constants==0)[0]])
    
    # exclude "Configuration Data XML" column by default
    if not arguments.include_configuration_XML:

        # remove "Configuration DATA XML" from headers
        try:
            const_cols.remove("Configuration Data XML")
        except:
            pass
    
        try:
            var_cols.remove("Configuration Data XML")
        except:
            pass

    # construct metadata table header
    meta_column_names = ['Part', 'Lot', 'Batch', 'Serial Number', 'Source'] + const_cols
    for tmds_type in metadata[0]["tdms_types"]:
        meta_column_names = meta_column_names + \
            [header + " [" + tmds_type + "]" for header in var_cols]
    
    # construct metadata table rows
    meta_rows = []
    for row in range(len(metadata)):

        # data from file names
        meta_row = [metadata[row]["part"], metadata[row]["lot"], metadata[row]["batch"],
            metadata[row]["sn"], metadata[row]["source"]]
        
        # data from tdms files
        root_props = metadata[row]["root_properties"]

        # constant root properties        
        meta_row = meta_row + list(root_props[0].iloc[0][const_cols].values)

        # variable root properties
        for i in range(len(root_props)):
            meta_row = meta_row + list(root_props[i].iloc[0][var_cols].values)

        # add to table
        meta_rows.append(meta_row)

    # log results
    log("Collated metadata table with " + str(len(meta_column_names)) + " columns and " +
        str(len(meta_rows)) + " rows.")

    return meta_column_names, meta_rows

# construct units table
def read_units_table (metadata, arguments, log):

    # header is always the same for DAC units table
    meta_var_names = ['Name', 'Time Units', 'Units', 'Plot Type']

    # fill in variables.meta table
    meta_vars = []
    for tdms_type in range(len(metadata[0]["tdms_types"])):
        for run_chart in metadata[0]["run_chart_headers"][tdms_type]:

            # name of run chart
            rc_name = run_chart + " [" + metadata[0]["tdms_types"][tdms_type] + "]"

            # units, default inferred
            units = "Not Given"
            if not arguments.do_not_infer_chart_units:
                units = RUN_CHART_UNITS[RUN_CHART_MATCHES.index(run_chart)]
            
            # construct row of table
            meta_vars.append([rc_name, 'Run', units, 'Curve'])

    # note if we inferred units
    if not arguments.do_not_infer_chart_units:
        log('Inferred run chart units from names.')

    return meta_var_names, meta_vars

# construct time step vectors
def read_timesteps (metadata):

    # go through data and find maximum length of each run chart
    time_steps_max = []
    for row in range(len(metadata)):

        # gather time steps per row
        time_steps_row = []
        for tdms_type in range(len(metadata[0]["tdms_types"])):
            time_steps_row = time_steps_row + \
                [metadata[row]["run_charts"][tdms_type].shape[0]] * \
                len(metadata[0]["run_chart_headers"][tdms_type])

        # get maximum time steps for all rows
        if row == 0:
            time_steps_max = time_steps_row
        else:
            time_steps_max = np.maximum(time_steps_max, time_steps_row)
  
    # construct time steps matrix
    time_steps = []
    for max in time_steps_max:
        time_steps.append(list(np.asarray(range(max)) + 1))

    return time_steps

# construct variable matrices
def read_variable_matrices (metadata, time_steps, arguments, log):

    
    # put maximum number of time steps
    # in format of run chart data
    max_time_steps = []
    k = 0
    for i in range(len(metadata[0]["tdms_types"])):
        run_chart_max_time_steps = []
        for j in range(len(metadata[0]["run_chart_headers"][i])):
            run_chart_max_time_steps.append(len(time_steps[k]))
            k = k + 1
        max_time_steps.append(run_chart_max_time_steps)

    # variable matrices are a list of matrices, one for each variable
    # rows are variable values, columns are time
    var_data = []
    inferred_variables = False
    for tdms_type in range(len(metadata[0]["tdms_types"])):
        for run_chart in range(len(metadata[0]["run_chart_headers"][tdms_type])):
            variable_i = []
            for row in range(len(metadata)):
                
                # get run chart data
                run_chart_data = list(metadata[row]["run_charts"][tdms_type].iloc[:,run_chart])

                # check for NaN values
                if np.isnan(run_chart_data).any():
                    log('Skipping NaN values in file "' + metadata[row]["tdms_files"][tdms_type] +
                        '" run chart "' + metadata[row]["run_charts"][tdms_type].columns[run_chart] + '".')
                    
                    # skip NaN values (we already checked that there 
                    # are some non-nan values when first reading tdms files)
                    run_chart_data = [x for x in run_chart_data if not np.isnan(x)]
                
                # extend run chart data by last value
                if len(run_chart_data) < max_time_steps[tdms_type][run_chart]:
                    if arguments.infer_last_value:
                        run_chart_data = run_chart_data + [run_chart_data[-1]] * \
                            (max_time_steps[tdms_type][run_chart] - len(run_chart_data))
                        inferred_variables = True
                    else:
                        raise TDMSUploadError('Found run chart length discrepancy in file "' +
                            metadata[row]["tdms_files"][tdms_type] + '".  Use inference options' +
                            ' (e.g. --infer-last-value) to correct lengths by inferring data.')
    
                # add run chart to current variable
                variable_i.append(run_chart_data)

            # add variable matrix to list
            var_data.append(variable_i)

    if inferred_variables:
        log("Variables inferred to correct run chart length discrepancies.")

    return var_data

# compute PCA for variables
def compute_PCA (var_data, arguments, log):

    # create pairwise distance matrix, using PCA or landmarks if available
    var_dist = []
    for i in range(len(var_data)):

        # compute PCA using sklearn
        pca = PCA(n_components=arguments.num_PCA_comps)
        try:
            dist_i = pca.fit_transform(var_data[i])
        except ValueError:
            raise TDMSUploadError("Could not perform PCA, too few components.")

        # save PCA projections
        var_dist.append(dist_i)

    return var_dist

# write data to given output directory
def write_dac_gen(meta_col_names, meta_rows, meta_var_names, meta_vars, 
            time_steps, var_data, var_dist, arguments):

    # create output directory if it doesn't already exist
    if os.path.exists(arguments.output_zip_file):
        log('Warning: overwriting .zip file "' + arguments.output_zip_file + '".')

    # write out zip file
    with ZipFile(arguments.output_zip_file, "w") as zip_file:

        # convert metadata table to string
        metadata_table = ','.join(meta_col_names) + "\n" + mat2str(meta_rows)

        # write out metadata table as "datapoints.dac"
        zip_file.writestr("datapoints.dac", metadata_table)

        # convert units table to string
        units_table = ','.join(meta_var_names) + '\n' + mat2str(meta_vars)

        # write out variables units table
        zip_file.writestr(os.path.join("var","variables.meta"), units_table)
        
        # write out pca.csv file with number of components
        zip_file.writestr("pca.csv", str(arguments.num_PCA_comps))
        
        # write out times/variables/distances
        num_vars = len(time_steps)
        for i in range(num_vars):

            # convert time steps to string
            time_i = time_steps[i]
            time_i = ','.join([str(time_i[j]) for j in range(len(time_i))])

            # write out to zip file as time/variable_i
            zip_file.writestr(os.path.join("time","variable_" + str(i+1) + ".time"), time_i)

            # write out variable data to zip file
            zip_file.writestr(os.path.join("var","variable_" + str(i+1) + ".var"), 
                mat2str(var_data[i]))

            # write out variable distances to zip file
            zip_file.writestr(os.path.join("dist","variable_" + str(i+1) + ".dist"), 
                mat2str(var_dist[i]))

    log('Output written to .zip file "' + arguments.output_zip_file + '".')

# helper function to convert a matrix to a comma separatedstring
def mat2str (mat):

    return '\n'.join([','.join([str(mat[j][k]).replace('\n', ' ').replace('\r', '').strip()
        for k in range(len(mat[j]))])
        for j in range(len(mat))])

# check arguments and create model
def create_model(arguments, log):

    # can't have both overvoltage and sprytron
    if arguments.overvoltage and arguments.sprytron:
        raise TDMSUploadError("Can't use both overvoltage and sprytron options " +
              "together. Please select one or the other and try again.")

    # check if zip file is correctly
    if not arguments.output_zip_file.endswith(".zip"):
        raise TDMSUploadError ('Must use .zip extension on output file name.')

    # organize tdms files
    metadata = catalog_tdms_files(arguments, log)

    # screen columns according to user input
    run_chart_matches = RUN_CHART_MATCHES
    if arguments.use_run_charts is not None:
        run_chart_matches = arguments.use_run_charts

    # report on run charts being used
    log("Using run charts in: " + str(run_chart_matches))

    # read tdms data
    metadata = read_tdms_files(metadata, run_chart_matches, log)

    # construct metadata table
    meta_col_names, meta_rows = read_metadata_table(metadata, arguments, log)

    # construct units table
    meta_var_names, meta_vars = read_units_table(metadata, arguments, log)

    # construct time matrix
    time_steps = read_timesteps (metadata)
    
    # construct variable matrices
    var_data = read_variable_matrices (metadata, time_steps, arguments, log)

    # compute PCA representation
    var_dist = compute_PCA (var_data, arguments, log)
    
    # save files
    write_dac_gen(meta_col_names, meta_rows, meta_var_names, meta_vars, 
                  time_steps, var_data, var_dist, arguments)

    # add output file for dac_gen script
    dac_gen_args = Namespace(**vars(arguments), **{'dac_gen_zip': arguments.output_zip_file})

    # push model using dac_gen
    if not arguments.do_not_upload:
        mid = dac_gen.upload_model(dac_gen_args, log)

        # supply the user with a direct link to the new model.
        host = arguments.host
        if arguments.port:
            host = host + ":" + arguments.port
        log("Your new model is located at %s/models/%s" % (host, mid))
        log('***** DAC Model Successfully Created *****')

    # should we erase the .zip file created
    if arguments.clean_up_output:
        os.remove(arguments.output_zip_file)
        log('Deleted file "' + arguments.output_zip_file + '".')

# logging is just printing to the screen
def log (msg):
    print(msg)

# set up argument parser
def parser ():

    # provide additional command line arguments for TDMS files
    parser = slycat.web.client.ArgumentParser(description=
        "Creates a Dial-A-Cluster model using run charts from .tdms files.")

    # input data directory
    parser.add_argument("input_data_dir", 
        help='Directory containing .tdms output files, organized ' +
             'according to part number.')

    # part number specification
    parser.add_argument("part_num_match",
        help='Part number to match when creating run chart model, can included wildcards, ' +
             'e.g. "XXXXXX*", or "XXXXXX_XX*".  Note you should use "" in Unix to pass wildcards.')

    # output file for dac generic format, mandatory
    parser.add_argument("output_zip_file",
        help="Output file name for writing dac generic format of the run " +
             " chart model.  If it already exists it will be overwritten.")

    # delete output file after successful model creation
    parser.add_argument("--clean-up-output", action="store_true",
        help="Delete output .zip file after successful model creation).")

    # do not upload to slycat
    parser.add_argument("--do-not-upload", action="store_true",
        help="Do not upload to slycat server.")

    # model and project names/descriptions
    parser.add_argument("--marking", default="ouo3", 
        help="Marking type.  Default: %(default)s")
    parser.add_argument("--model-description", default="", 
        help="New model description.  Default: %(default)s")
    parser.add_argument("--model-name", default="DAC Run Chart Model", 
        help="New model name.  Default: %(default)s")
    parser.add_argument("--project-description", default="", 
        help="New project description.  Default: %(default)s")
    parser.add_argument("--project-name", default="DAC Run Chart Models", 
        help="New project name.  Default: %(default)s")

    # spyrton/overvoltage parameters
    parser.add_argument("--overvoltage", action="store_true",
        help="Expecting overvoltage data.")
    parser.add_argument("--sprytron", action="store_true",
        help="Expecting sprytron data.")

    # run-chart file selection
    parser.add_argument('--tdms-file-matches', nargs="+", choices=TDMS_MATCHES,
        help='Use specified file matches to filter TDMS files, ' + 
             'defaults to full list of choices.')
    parser.add_argument("--use-run-charts", nargs="+", choices=RUN_CHART_MATCHES,
        help="Use specified variables for run-charts, otherwise run-charts " + 
             "are inferred from non-constant values.")

    # inferring data
    parser.add_argument("--do-not-infer-chart-units", action="store_true",
        help="Do not infer run chart units. Default is to infer units " +
             "from variable name.")
    parser.add_argument("--infer-last-value", action="store_true", 
        help="Fill out short run charts with last given value.")

    # exclude XML by default
    parser.add_argument("--include_configuration_XML", action="store_true",
        help='Include "Configuration Data XML" if present in data.')

    # compute using PCA model
    parser.add_argument("--num-PCA-comps", default=10, type=int,
        help="Number of PCA components to use, integer >= 2.  " +
             "Default: %(default)s")

    return parser

# tdms entry point
def main():

    # set up argument parser
    run_chart_parser = parser()  

    # get user arguments
    arguments = run_chart_parser.parse_args()

    # check arguments and create model
    create_model(arguments, log)

# command line version to load a single DAC model
if __name__ == "__main__":

    main()


# Copyright (c) 2013, 2018 National Technology and Engineering Solutions of Sandia, LLC . Under the terms of Contract
# DE-NA0003525 with National Technology and Engineering Solutions of Sandia, LLC, the U.S. Government
# retains certain rights in this software.

"""Slycat makes extensive use of `darray` objects - dense, multi-dimension,
multi-attribute arrays - as its fundamental unit of storage and organization.
In the abstract, a darray can be modeled as follows:

* A set of dimensions.  Each dimension has a name, index type, and a half-open range of valid index values.  Currently, the only supported index type is "int64", and indices are all zero-based (i.e. the range always begins at zero), but these may change in the future.  Collectively, the dimensions define the size and shape of the array.
* A set of attributes, each with a name and type.  Allowed attribute types include a full complement of signed and unsigned fixed-width integer types, plus floating-point and string types.  Collectively, attributes define *what* will be stored in the array.
* The array data.  Because darrays are dense, the data will include one value per attribute, for every location in the array.

This definition allows darrays to be flexible and efficient - for example, a
"table" data structure with heterogenous column types can be stored as a 1D
darray with multiple attributes, while a "matrix" would be stored as a 2D darray
with a single floating-point attribute.

Note that darrays are an abstract concept with multiple concrete
representations.  This module defines an abstract interface for manipulating
Python darrays, and a concrete implementation with in-memory storage.  The
:py:mod:`slycat.hdf5` module defines functionality for manipulating darrays
stored in HDF5 files on disk, and the :ref:`rest-api` defines functionality
for working with darrays using HTTP.

Note that it is rare to manipulate entire darrays in memory at once, due to
their size - most applications will work with *slices* of a darray to keep
memory use manageable.
"""

import numpy
import cherrypy


class Prototype(object):
  """Abstract interface for all darray implementations."""
  @property
  def ndim(self):
    """Return the number of dimensions in the array."""
    raise NotImplementedError()

  @property
  def shape(self):
    """Return the shape (size along each dimension) of the array."""
    raise NotImplementedError()

  @property
  def size(self):
    """Return the size (total number of elements) of the array."""
    raise NotImplementedError()

  @property
  def dimensions(self):
    """Return a description of the array dimensions."""
    raise NotImplementedError()

  @property
  def attributes(self):
    """Return a description of the array attributes."""
    raise NotImplementedError()

  def get_statistics(self, attribute=0):
    """Return statistics describing one attribute."""
    raise NotImplementedError()

  def get_data(self, attribute=0):
    """Return data from one attribute."""
    raise NotImplementedError()

  def set_data(self, attribute, slice, data):
    """Write data to one attribute."""
    raise NotImplementedError()

class Stub(Prototype):
  """darray implementation that only stores array metadata (dimensions and attributes)."""
  def __init__(self, dimensions, attributes):
    if len(dimensions) < 1:
      cherrypy.log.error("darray.py Stub.__init__", "At least one dimension is required.")
      raise ValueError("At least one dimension is required.")
    if len(attributes) < 1:
      cherrypy.log.error("darray.py Stub.__init__", "At least one attribute is required.")
      raise ValueError("At least one attribute is required.")

    self._dimensions = [dict(name=_require_dimension_name(dimension["name"]), type=_require_dimension_type(dimension.get("type", "int64")), begin=_require_dimension_bound(dimension.get("begin", 0)), end=_require_dimension_bound(dimension["end"])) for dimension in dimensions]
    self._attributes = [dict(name=_require_attribute_name(attribute["name"]), type=_require_attribute_type(attribute["type"])) for attribute in attributes]

    for dimension in self._dimensions:
      if dimension["begin"] != 0:
        cherrypy.log.error("darray.py Stub.__init__", "Dimension range must being with 0.")
        raise ValueError("Dimension range must begin with 0.")

  @property
  def ndim(self):
    """Return the number of dimensions in the array."""
    return len(self._dimensions)

  @property
  def shape(self):
    """Return the shape (size along each dimension) of the array."""
    return tuple([dimension["end"] - dimension["begin"] for dimension in self._dimensions])

  @property
  def size(self):
    """Return the size (total number of elements) of the array."""
    return numpy.prod(self.shape)

  @property
  def dimensions(self):
    """Return a description of the array dimensions."""
    return self._dimensions

  @property
  def attributes(self):
    """Return a description of the array attributes."""
    return self._attributes

class MemArray(Stub):
  """darray implementation that holds the full array contents in memory."""
  def __init__(self, dimensions, attributes, data):
    Stub.__init__(self, dimensions, attributes)

    if len(attributes) != len(data):
      cherrypy.log.error("darray.py MemArray.__init__", "Attribute and data counts must match.")
      raise ValueError("Attribute and data counts must match.")

    self._data = [numpy.array(attribute) for attribute in data]

    for attribute in self._data:
      if attribute.shape != self.shape:
        cherrypy.log.error("darray.py MemArray.__init__", "Attribute data must match array shape.")
        raise ValueError("Attribute data must match array shape.")

  def get_statistics(self, attribute=0):
    """Return statistics describing one attribute."""
    attribute = self._data[attribute]

    if attribute.dtype.char in ["O", "S", "U"]:
      return dict(min=min(attribute), max=max(attribute))

    attribute = attribute[numpy.invert(numpy.isnan(attribute))]
    if len(attribute):
      return dict(min=attribute.min(), max=attribute.max())

    return dict(min=None, max=None)

  def get_data(self, attribute=0):
    """Return a data slice from one attribute."""
    return self._data[attribute]

  def set_data(self, attribute, slice, data):
    """Write a data slice to one attribute."""
    self._data[attribute][slice] = data

def _require_attribute_name(name):
  if not isinstance(name, str):
    cherrypy.log.error("darray.py _require_attribute_name", "Attribute name must be a string.")
    raise ValueError("Attribute name must be a string.")
  return name

def _require_attribute_type(type):
  if type not in _require_attribute_type.allowed_types:
    cherrypy.log.error("darray.py _require_attribute_type", "Attribute type must be one of %s" % ",".join(_require_attribute_type.allowed_types))
    raise ValueError("Attribute type must be one of %s" % ",".join(_require_attribute_type.allowed_types))
  return type
_require_attribute_type.allowed_types = set(["int8", "int16", "int32", "int64", "uint8", "uint16", "uint32", "uint64", "float32", "float64", "string", "bool"])

def _require_dimension_name(name):
  if not isinstance(name, str):
    cherrypy.log.error("darray.py _require_attribute_name", "Dimension name must be a string.")
    raise ValueError("Dimension name must be a string.")
  return name

def _require_dimension_type(type):
  if type not in _require_dimension_type.allowed_types:
    cherrypy.log.error("darray.py _require_dimension_type", "Dimension type must be one of %s" % ",".join(_require_dimension_type.allowed_types))
    raise ValueError("Dimension type must be one of %s" % ",".join(_require_dimension_type.allowed_types))
  return type
_require_dimension_type.allowed_types = set(["int64"])

def _require_dimension_bound(bound):
  if not isinstance(bound, int) and type(bound) is not numpy.int64:
    cherrypy.log.error("darray.py _require_dimension_bound", "Dimension bound must be an integer.")
    raise ValueError("Dimension bound must be an integer.")
  return bound




# Copyright (c) 2013, 2018 National Technology and Engineering Solutions of Sandia, LLC . Under the terms of Contract
# DE-NA0003525 with National Technology and Engineering Solutions of Sandia, LLC, the U.S. Government
# retains certain rights in this software.

# This script lists the markings supported by the Slycat server.

import slycat.web.client
import numpy

# call to server to get markings and display results
def main(connection):

    # get markings
    markings = connection.get_configuration_markings()

    # output markings
    type_width = numpy.max([len(marking["type"]) for marking in markings])
    label_width = numpy.max([len(marking["label"]) for marking in markings])
    print()
    print("{:>{}} {:<{}}".format("Marking", type_width, "Description", label_width))
    print("{:>{}} {:<{}}".format("-" * type_width, type_width, "-" * label_width, label_width))
    for marking in markings:
        print("{:>{}} {:<{}}".format(marking["type"], type_width, marking["label"], label_width))
    print()

# command line entry point
if __name__ == "__main__":

    # get arguments for connecting to Slycat server
    parser = slycat.web.client.ArgumentParser(
        description="Display available marking types supported by the Slycat server.")
    arguments = parser.parse_args()

    # connect and get markings
    connection = slycat.web.client.connect(arguments)

    # call server to get markings
    main(connection)


# Copyright (c) 2013, 2018 National Technology and Engineering Solutions of Sandia, LLC . Under the terms of Contract
# DE-NA0003525 with National Technology and Engineering Solutions of Sandia, LLC, the U.S. Government
# retains certain rights in this software.

# This script read and processes a dac generic .zip file before pushing
# it to the Slycat server

# S. Martin
# 10/24/22

# file checks
import os

# connect to slycat server
import slycat.web.client

# DAC UI defaults
from slycat.web.client.dac_defaults import dac_model_defaults

# DAC error handling
class DACUploadError (Exception):

    # exception for DAC upload problems
    def __init__(self, message):
        self.message = message

# parse command line arguments
def parser():

    parser = slycat.web.client.ArgumentParser(description=
        "Create Slycat Dial-A-Cluster model from dac generic .zip file.")

    # input file
    parser.add_argument("dac_gen_zip", help="DAC generic .zip file.")

    # model and project names/descriptions
    parser.add_argument("--marking", default="cui", 
        help="Marking type.  Default: %(default)s")
    parser.add_argument("--model-description", default="", 
        help="New model description.  Default: %(default)s")
    parser.add_argument("--model-name", default="DAC Model", 
        help="New model name.  Default: %(default)s")
    parser.add_argument("--project-description", default="", 
        help="New project description.  Default: %(default)s")
    parser.add_argument("--project-name", default="DAC Models", 
        help="New project name.  Default: %(default)s")
        
    return parser

# logging is just printing to the screen
def log (msg):
    print(msg)

# create DAC model, show progress by default, arguments 
# include the command line connection parameters
# and project/model information
# create TDMS model, show progress by default, arguments 
# include the command line connection parameters
# and project/model information
def upload_model (arguments, log, progress=True):

    # setup a connection to the Slycat Web Server.
    connection = slycat.web.client.connect(arguments)

    # create a new project to contain our model.
    pid = connection.find_or_create_project(arguments.project_name, arguments.project_description)

    # create the new, empty model.
    mid = connection.post_project_models(pid, "DAC", arguments.model_name, arguments.marking, 
        arguments.model_description)

    # assign model UI defaults
    connection.put_model_parameter(mid, "dac-ui-parms", dac_model_defaults())
    
    # upload data to new model
    connection.upload_files(mid, [arguments.dac_gen_zip], 
        "dac-gen-zip-parser", [["Null"], ["DAC"]], progress)

    # mark model as finished
    connection.post_model_finish(mid)

    # wait until the model is ready
    connection.join_model(mid)

    return mid

# check arguments and create model
def create_model (arguments, log):

    # check that file is .zip file
    if not arguments.dac_gen_zip.endswith('.zip'):
        raise DACUploadError ('Must supply a DAC generic format file ending with ".zip".')
    
    # check that file exists
    if not os.path.isfile(arguments.dac_gen_zip):
        raise DACUploadError ('File "' + arguments.dac_gen_zip + '" does not exist.')

    # echo back user input, starting with .zip file
    log('*********** Creating DAC Model ***********')
    log("Input file: " + arguments.dac_gen_zip)

    # upload model
    mid = upload_model (arguments, log)

    # supply direct link to model
    host = arguments.host
    if arguments.port:
        host = host + ":" + arguments.port
    log("Your new model is located at %s/models/%s" % (host, mid))
    log('***** DAC Model Successfully Created *****')

# main entry point
def main():

    # set up argument parser
    ps_parser = parser()  

    # get user arguments
    arguments = ps_parser.parse_args()

    # check arguments and create model
    create_model(arguments, log)

# command line version
if __name__ == "__main__":

    main()


# Copyright (c) 2013, 2018 National Technology and Engineering Solutions of Sandia, LLC. 
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering 
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this software.

# This script creates a Python distribution wheel for the Slycat
# web client.  When run, it copies the latest version of slycat.web.client
# into the slycat_web_client directory 
#
# S. Martin
# 10/23/2020

# To publish to PyPi, perform the following steps:
#
# $ rm -rf dist
# $ python setup.py sdist bdist_wheel
# $ twine upload dist/*
#
# To publish to testpypi, use:
#
# $ twine upload --repository-url https://test.pypi.org/legacy/ dist/*
#
# The first step builds the distribution, and the second step
# uploads to PyPi.  To install the package from another computer use:
#
# $ pip install slycat-web-client
#
# To install from testpypi, use:
#
# $ pip install --extra-index-url https://testpypi.python.org/pypi slycat-web-client 
#   --upgrade --trusted-host testpypi.python.org --proxy wwwproxy.sandia.gov:80
#
# NOTE: Any local changes to slycat/web/client/__init__.py will be overwritten
# when you run this command!  Changes should be made to pacakges/slycat/web/client

from shutil import copyfile

# copy slycat.web.client and slycat.darray into slycat_web_client directory. This
# makes the slycat_web_directory a Python package without other Slycat dependencies.
copyfile('../packages/slycat/web/client/__init__.py', 'slycat/web/client/__init__.py')
copyfile('../packages/slycat/darray.py', 'slycat/darray.py')

# also copy the __init__.py files, which include the Slycat version number
copyfile('../packages/slycat/__init__.py', 'slycat/__init__.py')
copyfile('../packages/slycat/web/__init__.py', 'slycat/web/__init__.py')

# get Slycat version
import slycat
VERSION = slycat.__version__

# development version
# VERSION = VERSION
VERSION = "3.4.0"

# get README.md
import pathlib

# directory containing this file
HERE = pathlib.Path(__file__).parent

# text of the web-client-readme.txt file
README = (HERE / "README.md").read_text()

# create distribution
import setuptools

# create Python distribution wheel
from setuptools import setup

setup(
    name="slycat-web-client",
    version=VERSION,
    description="Slycat web client utilties for interacting with the Slycat " +
                "data analysis and visualization server.",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/sandialabs/slycat",
    author="Shawn Martin",
    author_email="smartin@sandia.gov",
    license="Sandia",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
    ],
    packages=setuptools.find_packages(),
    include_package_data=True,
    install_requires=["requests", "requests-kerberos",
                      "numpy", "cherrypy", "sklearn", 
                      "nptdms", "pandas"],
    entry_points={
        "console_scripts": [
            "dac_tdms=slycat.web.client.dac_tdms:main",
            "dac_tdms_batch=slycat.web.client.dac_tdms_batch:main",
            "ps_csv=slycat.web.client.ps_csv:main",
            "dac_gen=slycat.web.client.dac_gen:main"
        ]
    },
)
